import numpy as np
import geopandas as gpd
import pandas as pd
from datetime import datetime

from ..processing import read_raster
from ..processing import normalised_difference
from ..processing import mask_raster
from ..processing import autocategorize1D
from ..processing import px_count
from ..processing import colorize_raster
from ..processing import save_table


def forest_monitoring(image_name, aoi_mask, storage):
    # read_raster image
    ds, raster = read_raster(image_name, storage)

    date_and_tif = image_name.split("_")[-1]
    date = date_and_tif.split(".")[0]

    # calculate ndvi
    ndvi = normalised_difference(raster)

    # save_raster ndvi
    raster_name_ndvi = f"ndvi_{date}.tif"
    storage.create(ndvi, raster_name_ndvi, ds=ds)

    # read_geojson aoi_mask
    aoi_mask_gdf = gpd.GeoDataFrame.from_features(aoi_mask, crs=4326)

    # mask_raster ndvi with aoi_mask
    ndvi_masked, _ = mask_raster(raster_name_ndvi, aoi_mask_gdf, storage)

    # save_raster ndvi_masked
    raster_name_ndvi_masked = f"ndvi_masked_{date}.tif"
    storage.create(ndvi_masked, raster_name_ndvi_masked, ds=ds)

    # autocategorize1D ndvi
    ndvi_categorized = autocategorize1D(ndvi)

    # save_raster ndvi_categorized
    raster_name_ndvi_categorized = f"ndvi_categorized_{date}.tif"
    storage.create(ndvi_categorized, raster_name_ndvi_categorized, ds=ds)

    # apply_threshold to ndvi_categorized
    threshold = 3
    vegetation = ndvi_categorized >= threshold
    vegetation = vegetation.astype(int)

    # save_raster vegetation
    raster_name_vegetation = f"vegetation_{date}.tif"
    storage.create(vegetation, raster_name_vegetation, ds=ds)

    # mask_raster vegetation with aoi_mask
    vegetation_masked, _ = mask_raster(
        raster_name_vegetation, aoi_mask_gdf, storage
    )

    # save_raster vegetation_masked
    raster_name_vegetation_masked = f"vegetation_masked_{date}.tif"
    storage.create(vegetation_masked, raster_name_vegetation_masked, ds=ds)

    # colorize_raster vegetation_masked
    vegetation_masked_rgb = colorize_raster(vegetation_masked)

    # save_raster vegetation_masked_rgb
    raster_name_vegetation_masked_rgb = f"vegetation_masked_rgb_{date}.tif"
    storage.create(vegetation_masked_rgb, raster_name_vegetation_masked_rgb, ds=ds)

    # mask_raster ndvi_categorized with aoi_mask
    quality_mask, _ = mask_raster(
        raster_name_ndvi_categorized, aoi_mask_gdf, storage
    )

    # save_raster quality_mask
    raster_name_quality_mask = f"quality_masked_{date}.tif"
    storage.create(quality_mask, raster_name_quality_mask, ds=ds)

    # colorize_raster quality_mask
    quality_mask_rgb = colorize_raster(
        quality_mask, colors=["orangered", "yellow", "lawngreen", "darkgreen"]
    )

    # save_raster quality_mask_rgb
    raster_name_quality_mask_rgb = f"quality_masked_rgb_{date}.tif"
    storage.create(quality_mask_rgb, raster_name_quality_mask_rgb, ds=ds)

    # px_count vegetation_masked
    growth = px_count(vegetation_masked, values=[0, 1])

    # div growth
    growth_hectarias = np.divide(
        growth, 100, out=np.zeros_like(growth, dtype=np.float64), where=100 != 0
    )

    # save_table growth
    growth_table_name = "AOI_Vegetation_Growth.json"
    growth_columns = ["Not Vegetation Ha", "Vegetation Ha", "Total"]
    save_table(
        data=growth_hectarias,
        columns=growth_columns,
        table_name=growth_table_name,
        date=date,
        storage=storage,
    )

    # px_count quality_mask
    quality = px_count(quality_mask, values=[0, 1, 2, 3])

    # div quality
    quality_hectarias = np.divide(
        quality, 100, out=np.zeros_like(quality, dtype=np.float64), where=100 != 0
    )

    # save_table quality
    quality_table_name = "AOI_Vegetation_Quality.json"
    quality_columns = [
        "Bare Ground",
        "Sparse or Unhealthy Vegetation",
        "Healthy Vegetation",
        "Very Health Vegetation",
        "Total",
    ]
    save_table(
        data=quality_hectarias,
        table_name=quality_table_name,
        columns=quality_columns,
        date=date,
        storage=storage,
    )



from .SHDownloader import SHDownloader
from sentinelhub import MosaickingOrder


class SHS2Downloader(SHDownloader):
    def __init__(self, download_folder):
        super().__init__(download_folder)
        self.resolution = 10  # mpp
        self.mosaicking_order = MosaickingOrder.LEAST_CC
        self.script = """
            //VERSION=3
            function setup() {
                return {
                    input: [{
                        bands: ["B01","B02","B03","B04","B05","B06","B07","B08","B8A","B09","B11","B12"],
                        units: "DN"
                    }],
                    output: {
                        bands: 12,
                        sampleType: "INT16"
                    }
                };
            }

            function evaluatePixel(sample) {
                return [sample.B01,
                        sample.B02,
                        sample.B03,
                        sample.B04,
                        sample.B05,
                        sample.B06,
                        sample.B07,
                        sample.B08,
                        sample.B8A,
                        sample.B09,
                        sample.B11,
                        sample.B12];
            }
        """



from .read_raster import read_raster
from .normalised_difference import normalised_difference
from .mask_raster import mask_raster
from .autocategorize1D import autocategorize1D
from .px_count import px_count
from .colorize_raster import colorize_raster
from .save_table import save_table
from .rasterio_mask import mask_a_raster
from .convert_array_to_vector import convert_array_to_vector



import os
from glob import glob
import rasterio as rio
import numpy as np
import pandas as pd
import geopandas as gpd
import xarray as xr
import shutil

from .BaseStorage import BaseStorage


class Storage(BaseStorage):
    def __init__(self, path="data"):
        super().__init__()
        self.path = path
        if not os.path.exists(path):
            os.makedirs(path)
            print(f"{path} created")

    def exists(self, name):
        return os.path.exists(self.get_path(name))

    def get_path(self, name):
        return os.path.join(self.path, name)

    def create_from_path(self, data, name):
        dst_path = self.get_path(name)
        shutil.move(data, dst_path)
        return dst_path

    def create_from_image(self, data, name):
        dst_path = self.get_path(name)
        data.save(dst_path)
        return dst_path

    def create_from_rasterio(self, x, name, ds, window=None):
        dst_path = self.get_path(name)
        kwargs = ds.meta.copy()
        transform = ds.transform if window is None else ds.window_transform(window)
        kwargs.update(
            driver="GTiff",
            count=1 if x.ndim < 3 else x.shape[0],
            height=x.shape[0] if x.ndim < 3 else x.shape[1],
            width=x.shape[1] if x.ndim < 3 else x.shape[2],
            dtype=np.uint8 if x.dtype == "bool" else x.dtype,
            crs=ds.crs,
            transform=transform,
            # nbits=1 if x.dtype == 'bool' else
        )
        with rio.open(dst_path, "w", **kwargs) as dst:
            bands = 1 if x.ndim < 3 else [i + 1 for i in range(x.shape[0])]
            dst.write(x, bands)
        return dst_path

    def create_from_array(self, data, name):
        dst_path = self.get_path(name)
        np.save(dst_path, data)
        return dst_path

    def create_from_csv(self, data, name):
        dst_path = self.get_path(name)
        data.to_csv(dst_path)
        return dst_path

    def create_from_json(self, data, name):
        dst_path = self.get_path(name)
        data.to_json(dst_path)
        return dst_path
    
    def create_from_parquet(self, data, name):
        dst_path = self.get_path(name)
        data.to_parquet(dst_path)
        return dst_path
    
    def create_from_zarr(self, data, name):
        dst_path = self.get_path(name)
        data.to_zarr(dst_path)
        return dst_path

    def list(self, pattern="*"):
        paths = glob(os.path.join(self.path, pattern))
        # strip base path
        return [p.replace(self.path + "/", "") for p in paths]

    def read_from_array(self, name, path=None):
        if path is None:
            path = self.get_path(name)
        return np.load(path)

    def read_from_rasterio(self, name):
        return rio.open(self.get_path(name))

    def read_from_csv(self, name):
        return pd.read_csv(self.get_path(name), index_col=0)

    def read_from_json(self, name):
        return pd.read_json(self.get_path(name))

    def read_from_parquet(self, name):
        return gpd.read_parquet(self.get_path(name))
    
    def read_from_zarr(self, name):
        return xr.open_zarr(self.get_path(name))



import requests
import os
from rich import print


class APIRepo:
    def __init__(
        self, url=os.getenv("SPAI_API_URL", "http://spai.test.earthpulse.ai/")
    ):
        # def __init__(self, url=os.getenv("SPAI_API_URL", "http://localhost:8000/")):
        self.url = url
        self.url = self.url if self.url.endswith("/") else self.url + "/"
        if not self.url.startswith("http://") and not self.url.startswith("https://"):
            raise Exception(
                "Invalid URL: " + self.url + "\nIt must start with http:// or https://"
            )

    def format_response(self, response):
        # return (data, error)
        if response.status_code == 200:
            return response.json(), None
        return None, response.json()["detail"]

    def info(self):
        return requests.get(self.url).json()

    def login(self):
        return requests.get(self.url + "auth/login")

    def token(self, code):
        return requests.get(self.url + "auth/token?code=" + code)

    def logout_url(self):
        response = requests.get(self.url + "auth/logout")
        return response.json()["logout_url"]

    def get_headers(self, user):
        return {"Authorization": "Bearer " + user["id_token"]}

    def retrieve_projects(self, user):
        return requests.get(
            self.url + "projects", headers=self.get_headers(user)
        ).json()

    def retrieve_project(self, user, project):
        return requests.get(
            self.url + f"projects/{project}", headers=self.get_headers(user)
        )

    def retrieve_project_by_name(self, user, project_name):
        response = requests.get(
            self.url + f"projects?name={project_name}", headers=self.get_headers(user)
        )
        return self.format_response(response)

    def stop_service(self, user, service_id):
        response = requests.post(
            self.url + f"stop/{service_id}",
            headers=self.get_headers(user),
        )
        if response.status_code == 200:
            return response.text
        return "Something went wrong.\n" + response.json()["detail"]

    def stop_service_by_name(self, user, project_id, service_type, service_name):
        response = requests.post(
            self.url
            + f"stop?project={project_id}&type={service_type}&name={service_name}",
            headers=self.get_headers(user),
        )
        return self.format_response(response)

    def get_logs(self, user, service_id):
        response = requests.get(
            self.url + f"logs/{service_id}",
            headers=self.get_headers(user),
        )
        if response.status_code == 200:
            return response.json()
        return "Something went wrong.\n" + response.json()["detail"]

    def delete_project(self, user, project):
        response = requests.delete(
            self.url + f"projects/{project}", headers=self.get_headers(user)
        )
        return self.format_response(response)

    def retrieve_template(self, user, template):
        return requests.get(
            self.url + f"templates/{template}", headers=self.get_headers(user)
        )

    def deploy_template(self, user, template):
        response = requests.post(
            self.url + f"deploy/template?name={template}",
            headers=self.get_headers(user),
        )
        if response.status_code == 200:
            print(response.text)
            return
        print("[red]Something went wrong: " + response.json()["detail"])

    def deploy_folder(self, user, zip_path):
        # Send the request
        response = requests.post(
            self.url + f"deploy",
            headers=self.get_headers(user),
            files={"folder": open(zip_path, "rb")},
        )
        return self.format_response(response)

    def retrieve_service_by_name_type_project(
        self, user, project_id, service_type, service_name
    ):
        response = requests.get(
            self.url
            + f"services?project={project_id}&type={service_type}&name={service_name}",
            headers=self.get_headers(user),
        )
        return self.format_response(response)

    def retrieve_service(self, user, service_id):
        response = requests.get(
            self.url + f"services/{service_id}",
            headers=self.get_headers(user),
        )
        if response.status_code == 200:
            return response.json()
        print("[red]Something went wrong: " + response.json()["detail"])



"""
OpenStreetMapa data download module
"""
from typing import List, Optional, Union

import geopandas as gpd
import osmnx as ox
import pandas as pd
from shapely.geometry import Polygon
from shapely.validation import make_valid


def download_waterways_from_osm(
    gdf: Union[str, gpd.GeoDataFrame],
    storage,
    waterway_tags: Optional[dict] = {
        "waterway": ["river", "canal", "stream", "brook", "ditch", "drain"]
    },
) -> None:
    """
    Download the waterways from OpenStreetMap for the given area of interest

    Parameters
    ----------
    aoi : Union[str, gpd.GeoDataFrame]
        The area of interest
    storage : BaseStorage
        The storage object
    waterway_tags : dict, optional
        The waterway tags to use, by default {'waterway': ['river', 'canal', 'stream', 'brook', 'ditch', 'drain']}

    Raises
    ------
    TypeError
        If the area of interest is not a GeoDataFrame or a filepath
    """
    if isinstance(gdf, str):
        gdf = gpd.read_file(gdf)
    elif not isinstance(gdf, gpd.GeoDataFrame):
        raise TypeError("AoI must be a GeoDataFrame or a filepath")

    all_waterways = get_all_osm_elements(gdf, waterway_tags)

    final_waterways_gdf = pd.concat(all_waterways, ignore_index=True)
    final_waterways_gdf = final_waterways_gdf[["geometry"]]
    final_waterways_gdf = final_waterways_gdf[
        final_waterways_gdf.geometry.type.isin(("LineString", "MultiLineString"))
    ]
    storage.create_from_parquet(final_waterways_gdf, name="waterways.parquet")


def download_roads_from_osm(
    gdf: Union[str, gpd.GeoDataFrame],
    storage,
    waterway_tags: Optional[dict] = {
        "highway": ["motorway", "trunk", "primary", "secondary", "tertiary"]
    },
) -> None:
    """
    Download the waterways from OpenStreetMap for the given area of interest

    Parameters
    ----------
    aoi : Union[str, gpd.GeoDataFrame]
        The area of interest
    storage : BaseStorage
        The storage object
    waterway_tags : dict, optional
        The waterway tags to use, by default {'highway': ['motorway', 'trunk', 'primary', 'secondary', 'tertiary']}

    Raises
    ------
    TypeError
        If the area of interest is not a GeoDataFrame or a filepath
    """
    if isinstance(gdf, str):
        gdf = gpd.read_file(gdf)
    elif not isinstance(gdf, gpd.GeoDataFrame):
        raise TypeError("AoI must be a GeoDataFrame or a filepath")

    all_roads = get_all_osm_elements(gdf, waterway_tags)

    final_roads_gdf = pd.concat(all_roads, ignore_index=True)
    final_roads_gdf = final_roads_gdf[["geometry"]]
    final_roads_gdf = final_roads_gdf[
        final_roads_gdf.geometry.type.isin(("LineString", "MultiLineString"))
    ]
    storage.create_from_parquet(final_roads_gdf, name="roads.parquet")


def download_buildings_from_osm(gdf: Union[str, gpd.GeoDataFrame], storage):
    """
    Download the buildings from OpenStreetMap for the given area of interest

    Parameters
    ----------
    aoi : Union[str, gpd.GeoDataFrame]
        The area of interest
    storage : BaseStorage
        The storage object

    Raises
    ------
    TypeError
        If the area of interest is not a GeoDataFrame or a filepath
    """
    if isinstance(gdf, str):
        gdf = gpd.read_file(gdf)
    elif not isinstance(gdf, gpd.GeoDataFrame):
        raise TypeError("AoI must be a GeoDataFrame or a filepath")

    all_buildings = []

    for _, row in gdf.iterrows():
        polygon = row.geometry

        try:
            buildings_gdf = ox.features_from_polygon(polygon, tags={"building": True})
        except ox._errors.InsufficientResponseError:
            continue

        all_buildings.append(buildings_gdf)

    final_buildings_gdf = pd.concat(all_buildings, ignore_index=True)
    final_buildings_gdf = final_buildings_gdf[["geometry"]]
    final_buildings_gdf = final_buildings_gdf[
        final_buildings_gdf.geometry.type.isin(("Polygon", "MultyPolygon"))
    ]
    final_buildings_gdf.to_crs(epsg=4326, inplace=True)
    storage.create_from_parquet(final_buildings_gdf, name="buildings.parquet")


def get_all_osm_elements(gdf: gpd.GeoDataFrame, tags: dict) -> List:
    """
    Iterate over a given AOI GeoDataFrame and get all the desired OSM elements defined in the tags

    Parameters
    ----------
    gdf: gpd.GeoDataFrame
        The GeoDataFrame of the Area of Interest
    tags: dict
        Dict with the OSM tags of the elements to download

    Returns
    -------
    all_elements: List
        List with all the required elements
    """
    all_elements = []

    for _, row in gdf.iterrows():
        # create polygon from row as shapely.geometry.Polygon
        polygon = row.geometry
        if not polygon.is_valid:
            polygon = make_valid(polygon)

        for key, values in tags.items():
            if isinstance(values, list):
                for value in values:
                    try:
                        waterway_gdf = ox.features_from_polygon(
                            polygon, tags={key: value}
                        )
                        all_elements.append(waterway_gdf)
                    except ox._errors.InsufficientResponseError:
                        continue
            else:
                try:
                    waterway_gdf = ox.features_from_polygon(polygon, tags={key: values})
                    all_elements.append(waterway_gdf)
                except ox._errors.InsufficientResponseError:
                    continue

    return all_elements



import numpy as np


def normalised_difference(raster, bands=[1, 2]):
    # convertir bandas a float
    if raster.shape[-3] < 2:
        raise ValueError(
            f"Function `normalised_difference` strictly expects at least a 2-band image. "
            f"You provided a {raster.shape[-3]}-band image instead."
        )

    bands = np.array(bands) - 1
    # Separate the bands
    band1 = raster[bands[0], :, :]
    band2 = raster[bands[1], :, :]
    # convert the bands to floats
    band1 = band1.astype(float)
    band2 = band2.astype(float)
    # calculate the ndvi
    numerator = band1 - band2
    denominator = band1 + band2 + 1e-8

    return numerator / denominator



import os
import geopandas as gpd
import geojson
from warnings import warn
from shapely.geometry import box
from shapely.geometry.polygon import Polygon
import requests


def validate_coords(coords):
    # Coordinates of points of a polygon: [[lon1,lat1],[lon2,lat2], ... ,[lonN,latN]]
    for coord in coords:
        # check that each coord is a list of 2 coordinates
        if len(coord) != 2:
            raise Exception("each coord must be a list of 2 coordinates")
        # check that each coordinate is a float
        if not isinstance(coord[0], float) or not isinstance(coord[1], float):
            raise Exception("each coordinate must be a float")
        lon = coord[0]
        lat = coord[1]

        # check lat and long ranges
        if lon < -180 or lon > 180 or lat < -90 or lat > 90:
            raise Exception("each coordinate must be a valid lat/long")

    return coords


def validate_bounds(bounds):
    # check that bounds is a list
    if len(bounds) != 4 or not isinstance(bounds, tuple):
        raise Exception(
            "bounds must be a tuple of 4 points: (minlon, minlat, maxlon, maxlat))"
        )
    # check that each bound is a float
    for bound in bounds:
        if not isinstance(bound, float):
            raise Exception("each bound must be a float")
    # check lat and long ranges
    minlon = bounds[0]
    minlat = bounds[1]
    maxlon = bounds[2]
    maxlat = bounds[3]

    if minlat < -90 or minlat > 90 or maxlat < -90 or maxlat > 90:
        raise Exception("each latitude must be a valid lat/long")
    if minlon < -180 or minlon > 180 or maxlon < -180 or maxlon > 180:
        raise Exception("each longitude must be a valid lat/long")
    # check that minlat < maxlat and minlng < maxlng
    if minlon > maxlon or minlat > maxlat:
        raise Exception(
            "minlat must be less than maxlat and minlng must be less than maxlng"
        )
    return bounds


def create_aoi_geodataframe(obj):
    # obj can be a :
    # GeoDataFrame,
    # GeoJSON file,
    # Path to a GeoJSON file,
    # Polygon of Shapely (Bounding Box),
    # List of Coords of a (Bounding Box),
    # Location name (string)

    # return a GeoDataFrame of the Bounding Box of the obj or a list of gdfs if obj is a location name and there are more than one relation

    if (
        isinstance(obj, str)
        and not os.path.isfile(obj)
        and not obj.endswith(".geojson")
    ):
        # case 1: obj is a location name (e.g. "Madrid")
        obj = get_box_or_gdfs_by_place_name(
            obj
        )  # return Box of Shapely or list of gdfs
        if isinstance(obj, list):
            return obj

    elif isinstance(obj, list):
        # case 2: obj is a list of coords of a Polygon: [[lat1,long1],[lat2,long2],...,[latN,longN]]
        coords = validate_coords(obj)
        poly = Polygon(coords)
        bounds = poly.bounds
        obj = box(*bounds)  # return Box of Shapely

    if isinstance(obj, tuple):
        # case 3: obj is a tuple of bounds: (minlat, minlng, maxlat, maxlng)
        bounds = validate_bounds(obj)
        obj = box(*bounds)  # return Box of Shapely

    if isinstance(obj, Polygon):
        # case 4: obj is a Box of Shapely: Polygon[[lat1,long1],[lat2,long2],[lat3,long3],[lat4,long4]]
        bounds = obj.bounds
        validate_bounds(bounds)
        return gpd.GeoDataFrame(geometry=[obj], crs=4326)

    if isinstance(obj, gpd.GeoDataFrame):
        # case 5: obj is a GeoDataFrame
        if not obj.crs:
            warn("GeoDataFrame has no crs, assuming EPSG:4326")
            obj.set_crs(epsg=4326, inplace=True)
        return obj.to_crs(4326)

    if (
        isinstance(obj, dict)
        and obj.get("type") == "FeatureCollection"
        and "features" in obj
    ):
        # case 6: obj is a GeoJSON file
        if obj["type"] == "Polygon":
            for coords in obj["coordinates"][0]:
                validate_coords(coords)
        elif obj["type"] == "MultiPolygon":
            for _coords in obj["coordinates"][0]:
                for coords in _coords:
                    validate_coords(coords)
        return gpd.GeoDataFrame.from_features(obj, crs=4326)

    if isinstance(obj, str) and os.path.isfile(obj) and obj.endswith(".geojson"):
        # case 7: obj is a path to GeoJSON file
        geojson_file = geojson.load(open(obj))
        if geojson_file["type"] == "Polygon":
            for coords in geojson_file["coordinates"][0]:
                validate_coords(coords)
        elif geojson_file["type"] == "MultiPolygon":
            for _coords in geojson_file["coordinates"][0]:
                for coords in _coords:
                    validate_coords(coords)
        return gpd.GeoDataFrame.from_features(geojson_file, crs=4326)

    raise Exception(f"location {obj} not supported")


def get_bb_by_city_name(city_name):
    base_url = "https://nominatim.openstreetmap.org"
    format_out = "json"
    limit = 10

    # Construct the API request URL
    url = f"{base_url}/search?city={city_name}&format={format_out}&limit={limit}"

    # Send the API request
    response = requests.get(url).json()

    results = []
    if len(response) == 0:
        raise Exception("No results found")
    for result in response:
        if "boundingbox" in result:
            bounding_box = result["boundingbox"]
            if len(bounding_box) == 4:
                min_lon, max_lon, min_lat, max_lat = map(float, bounding_box)
                results.append(
                    {
                        "name": f"{result['display_name']}",
                        "bbox": box(min_lat, min_lon, max_lat, max_lon),
                    }
                )

    return results


def get_box_or_gdfs_by_place_name(place_name):
    if not isinstance(place_name, str):
        raise Exception("place_name must be a string")
    results = get_bb_by_city_name(place_name)
    gdfs = []
    if len(results) == 1:
        # Only one result, access the bounding box directly
        return results[0]["bbox"]
    else:
        # Multiple results, iterate over the dictionary
        for dict in results:
            gdf = gpd.GeoDataFrame(geometry=[dict["bbox"]], crs=4326)
            gdfs.append({"name": dict["name"], "gdf": gdf})
        return gdfs



from ..repos import APIRepo

from .get_project_by_name import get_project_by_name


def delete_project(user, project):
    repo = APIRepo()
    project = get_project_by_name(user, project)
    data, error = repo.delete_project(user, project["id"])
    if error:
        raise Exception("Something went wrong.\n" + error)
    return data



from pydantic import BaseModel


class StorageConfig(BaseModel):
    name: str = "storage"


class LocalStorageConfig(StorageConfig):
    type: str = "local"
    path: str = "data"


class S3StorageCredentials(BaseModel):
    url: str
    access_key: str
    secret_key: str
    region: str = "us-east-1"
    bucket: str


class S3StorageConfig(StorageConfig):
    credentials: S3StorageCredentials



import os
import json
import yaml
from ..models import Config
from rich import print
from typing import List


def validate_folder(dir, folder, typer):
    if not os.path.exists(dir / folder):
        raise typer.BadParameter(f"No {folder} directory found in '{dir}'.")


def validate_item(dir, folder, item, name, typer, file="main.py"):
    # check name
    if not item.name:
        raise typer.BadParameter(f"{name} '{item.name}' is missing 'name' attribute.")
    # check folder has folder with item name
    if not os.path.exists(dir / folder / item.name):
        raise typer.BadParameter(
            f"{name} '{item.name}' cannot be found in {dir}/{folder}."
        )
    # check folder has file
    if not file in os.listdir(dir / folder / item.name):
        raise typer.BadParameter(f"{name} '{item.name}' is missing file 'main.py'.")
    # TODO: check optionals: reqs, env...


def load_and_validate_config(dir, typer, verbose=False):
    # check dir exists
    if not dir.exists():
        raise typer.BadParameter(f"Directory '{dir}' does not exist.")
    # check dir is a spai project
    if not "spai.config.yaml" in os.listdir(dir):
        raise typer.BadParameter(
            f"Directory '{dir}' is not a spai project. No spai.config.yaml file found."
        )
    # load config
    config = {}
    with open(dir / "spai.config.yaml", "r") as f:
        config = yaml.safe_load(f)
    if not config:
        raise typer.BadParameter(f"spai.config.yaml file is empty.")
    config.update(dir=dir)
    if verbose:
        print(config)
    config = Config(**config)
    # TODO: check if project name is already taken in cloud, locally is not a problem
    config.project = dir.name if not config.project else config.project
    # check scripts
    if config.scripts:
        # check project has scripts folder
        validate_folder(dir, "scripts", typer)
        for script in config.scripts:
            validate_item(dir, "scripts", script, "script", typer)
    # check apis
    if config.apis:
        # check project has apis folder
        validate_folder(dir, "apis", typer)
        for api in config.apis:
            validate_item(dir, "apis", api, "api", typer)
    # check uis
    if config.uis:
        # check project has uis folder
        validate_folder(dir, "uis", typer)
        for ui in config.uis:
            validate_item(dir, "uis", ui, "ui", typer)
    # check notebooks
    if config.notebooks:
        # check project has notebooks folder
        validate_folder(dir, "notebooks", typer)
        for notebook in config.notebooks:
            validate_item(dir, "notebooks", notebook, "notebook", typer, "main.ipynb")
    return config


def validate_template_parameters(url: str, aoi: str, dates: List) -> bool:
    """
    Validate the parameters of the template command.

    Parameters
    ----------
    url : str
        URL of the template repository.
    aoi : str
        Path to the AOI GeoJSON file.
    dates : List
        Start and end dates of the time interval.

    Returns
    -------
    bool
        True if the parameters are valid, False otherwise.
    """
    if not url:
        raise BadParameter(f"Template URL must be provided")
    if not validators.url(url):
        raise BadParameter(f"Invalid url '{url}'")

    if not aoi:
        raise FileNotFoundError(f"AOI file must be provided")
    if not os.path.exists(aoi):
        raise FileNotFoundError(f"AOI file not found in {aoi}")

    if not dates:
        raise BadParameter(f"Dates must be provided")

    return True



import os

from ..config import load_and_validate_config


def install_requirements(path, typer, verbose):
    config = load_and_validate_config(path, typer, verbose)
    # install requirements.txt in each service folder
    for service in config:
        service_path = config.dir / config.type2folder(service.type) / service.name
        reqs_path = service_path / "requirements.txt"
        if not reqs_path.exists():
            # typer.echo(
            #     f"No requirements.txt found for {service.type} '{service.name}'."
            # )
            continue
        typer.echo(f"Installing requirements for {service.type} '{service.name}'...")
        os.system(f"pip install -r {str(reqs_path)}")
    return "done"



from ..repos import APIRepo


def get_project_by_name(user, project):
    repo = APIRepo()
    data, error = repo.retrieve_project_by_name(user, project)
    if data:
        return data
    raise Exception("Something went wrong.\n" + error)



import os
import zipfile
import shutil

from ..config import load_and_validate_config
from ..repos import APIRepo


def create_zip_from_folders(folders, files, zip_path, root):
    parent_root = os.path.dirname(root)
    with zipfile.ZipFile(zip_path, "w") as zipf:
        for folder in folders:
            for folderName, subfolders, filenames in os.walk(folder):
                for filename in filenames:
                    filePath = os.path.join(folderName, filename)
                    zipf.write(filePath, os.path.relpath(filePath, parent_root))
        for file in files:
            zipf.write(file, os.path.relpath(file, parent_root))


def deploy_folder(user, path, typer, verbose=False):
    config = load_and_validate_config(path, typer, verbose=verbose)
    typer.echo(f"Deploying...")
    # zip the folder
    folders, files = [], []
    for script in config.scripts:
        folders.append(f"{path}/scripts/{script.name}")
    for api in config.apis:
        folders.append(f"{path}/apis/{api.name}")
    files.append(f"{path}/spai.config.yaml")
    dst_path = f"/tmp/spai"
    os.makedirs(dst_path, exist_ok=True)
    zip_path = f"{dst_path}/{config.project}.zip"
    create_zip_from_folders(folders, files, zip_path, root=config.project)
    # send to api for deployment
    repo = APIRepo()
    data, error = repo.deploy_folder(user, zip_path)
    # clean up
    shutil.rmtree(dst_path)
    if data:
        typer.echo(
            # f"Check status at {os.getenv('SPAI_UI_URL', 'https://spai.earthpulse.ai')}/dashboard/{data}"
            f"Check status at {os.getenv('SPAI_UI_URL', 'https://localhost:5173')}/dashboard/{data}"
        )
    raise Exception(f"Error: {error}")



from .SHDownloader import SHDownloader
from sentinelhub import DataCollection


class SHDEM30Downloader(SHDownloader):

    def __init__(self, download_folder='/tmp'):
        super().__init__(download_folder)
        self.data_collection = DataCollection.DEM_COPERNICUS_30
        self.resolution = 30  # mpp
        self.script = """
            //VERSION=3

            function setup() {
                return {
                    input: ["DEM"],
                    output: { id: "default",
                            bands: 1,
                            sampleType: SampleType.FLOAT32
                    },
                }
            }

            function evaluatePixel(sample) {
                return [sample.DEM]
            }
        """



import requests
import os

from .APIRepo import APIRepo


class ReportsRepo(APIRepo):
    def generate(self, body):
        return requests.post(self.url + "reports/generate", json=body, stream=True)



"""
Module for report generation
"""
import json
from typing import Optional, Union

import pandas as pd

from ..repos.ReportsRepo import ReportsRepo


def generate_report(
    dataframe: Union[dict, pd.DataFrame],
    plot_type: str,
    output: Optional[str] = "report.pdf",
    title: Optional[str] = None,
    text: Optional[str] = None,
    show_table: Optional[bool] = False,
    zip: Optional[bool] = False,
) -> None:
    """
    Generate a report from a dataframe

    Parameters
    ----------
    dataframe : dict
        The dataframe to be used in the report in JSON format
    output : str
        The path to the output file

    Returns
    -------
    None
    """
    reports_repo = ReportsRepo()

    # Convert dataframe to JSON if it is not already
    if isinstance(dataframe, str):
        try:
            dataframe = json.loads(dataframe)
        except TypeError:
            dataframe = dataframe.to_json()
    elif isinstance(dataframe, pd.DataFrame):
        dataframe = dataframe.to_json()

    body = {
        "dataframe": dataframe,
        "title": title,
        "text": text,
        "plot_type": plot_type,
        "show_table": show_table,
        "zip": zip,
    }
    report_bytes = reports_repo.generate(body)
    with open(output, "wb") as file:
        file.write(report_bytes.content)



# from datetime import datetime, timedelta
# from dotenv import load_dotenv

# from spai.storage import Storage
# from spai.data.satellite import explore_satellite_images, download_satellite_image

# load_dotenv()

# # explore available images
# print("Looking for images in the last month")
# bbox = (2.0052191, 41.38238988, 2.17002873, 41.48530451)  # Collserola
# images = explore_satellite_images(bbox, cloud_cover=10)
# if len(images) == 0:
#     raise ValueError("No images found")

# # download images and save locally
# storage = Storage("data")
# sensor = "S2L2A"  # or 'S2L1C'
# existing_images = storage.list(f"{sensor}*.tif")
# dates = [image.split("_")[1].split(".")[0] for image in existing_images]
# print("Found", len(images), f"image{'s' if len(images) > 1 else ''}")
# for image in images:
#     # check if image is already downloaded
#     date = image["date"].split("T")[0]
#     if date in dates:
#         print("Image already downloaded:", date)
#         continue
#     print("Downloading new image:", date)
#     path = download_satellite_image(storage, bbox, date, sensor)
#     print("Image saved at", path)


print("hola")



import numpy


def autocategorize1D(raster, iterations=200, centers=[0.1, 0.35, 0.50, 0.70]):
    # cluster pixels in given classes
    iterations = max([1, min([iterations, 200])])
    shape = raster.shape
    raster = raster.reshape(-1, 1)
    for i in range(iterations):
        clusters = numpy.argmin(numpy.abs(raster - centers), axis=1)
        centers = [numpy.mean(raster[clusters == j]) for j in range(len(centers))]
    return clusters.reshape(shape)



from .sentinelhub import SHExplorer
from .utils import create_aoi_geodataframe
from datetime import datetime, timedelta


def explore_satellite_images(aoi, time_interval=None, sensor="S2L2A", **kargs):
    if time_interval is None:
        # last month
        now = datetime.now()
        last_months = now - timedelta(days=30)
        time_interval = (last_months.strftime("%Y-%m-%d"), now.strftime("%Y-%m-%d"))
    # TODO: validate time format
    gdf = create_aoi_geodataframe(aoi)
    if isinstance(gdf, list):
        return gdf
    explorer = SHExplorer(time_interval, sensor, **kargs)
    results = explorer.search(gdf)
    return results



import numpy as np


def px_count(raster, values=None):
    """
    Counts the number of pixels in a raster.

    Parameters
    ----------
    raster : numpy.ndarray
        A 2D numpy array representing a raster.
    values : list, optional
        A list of values to count. If None, the total number of pixels is counted.
        If empty, the number of unique values is counted. If not empty, the number
        of pixels for each value in the list is counted.
        
    """
    flattened_raster = raster.flatten()
    counts = np.bincount(flattened_raster.astype(np.int64))

    if values is None:
        total_count = np.sum(counts)
        return np.append(counts, total_count)
    elif len(values) == 0:
        non_zero_indexes = [index for index, value in enumerate(counts) if value != 0]
        unique_count = len(non_zero_indexes)
        return np.append(counts, unique_count)
    else:
        desired_counts = []
        for value in values:
            if value < len(counts):
                desired_counts.append(counts[value])
            else:
                desired_counts.append(0)
        total_count = np.sum(desired_counts)
        desired_counts.append(total_count)
        return np.array(desired_counts)



"""
Module for sending emails
"""
import base64
import logging
import warnings

from typing import List, Union
from os.path import basename, splitext

from .mimetypes import MIMETYPES
from ..errors import MailError
from ..repos.MailsRepo import MailsRepo

logger = logging.getLogger(__name__)

ACCEPTED_RESPONSES = (200, 201, 202)


def send_mail(
    to: Union[List, str],
    subject: str,
    body: str,
    attachment: Union[List, str, None] = None,
):
    """
    Send an email.

    Parameters
    ----------
    to : Union[str, List]
        The email address of the recipient
    subject : str
        The subject of the email
    body : str
        The body of the email
    attachment : Union[List, str, None], optional
        The attachment of the email, if needed, by default None
    """
    mails_repo = MailsRepo()

    body = {"to": to, "subject": subject, "body": body}  # TODO validate

    if attachment:
        if isinstance(attachment, str):
            attach = curate_attachment(attachment)
        elif isinstance(attachment, list):
            attach = [curate_attachment(path) for path in attachment]

        body["attachment"] = attach

    r = mails_repo.send(body)

    if r.status_code not in ACCEPTED_RESPONSES:
        m = f"Email could not be sent. Response: {r.status_code}; Reason: {r.reason}, {r.text}"
        raise MailError(m)

    return "Mail sent successfully"


def encode_file(path: str):
    """
    Encode a file to base64

    Parameters
    ----------
    path : str
        Path to the file to encode

    Returns
    -------
    str
        The encoded file
    """
    try:
        with open(path, "rb") as f:
            data = f.read()
            f.close()
    except FileNotFoundError:
        m = f"File {path} not found"
        logger.error(m)
        raise FileNotFoundError(m)

    encoded_file = base64.b64encode(data).decode()

    return encoded_file


def get_filetype(path: str):
    """
    Get the filetype of a file

    Parameters
    ----------
    path : str
        Path to the file

    Returns
    -------
    filetype
        The filetype of the file

    Raises
    ------
    Exception
        If the filetype cannot be guessed
    """
    ext = splitext(path)[1]

    if ext in MIMETYPES:
        filetype = MIMETYPES[ext]
    else:
        filetype = "application/octet-stream"
        m = f"MIMEType of {path} is not in MIMETYPES, assuming {filetype}"
        logger.warning(m)
        warnings.warn(m)

    return filetype


def curate_attachment(path: str):
    """
    Curate an attachment to be sent

    Parameters
    ----------
    path : str
        Path to the file to attach

    Returns
    -------
    dict
        The attachment to be sent
    """
    encoded_file = encode_file(path)
    filetype = get_filetype(path)
    filename = basename(path)

    return {"filename": filename, "encoded_file": encoded_file, "filetype": filetype}



from ..repos import APIRepo


def get_services(user, project_name):
    repo = APIRepo()
    data, error = repo.retrieve_project_by_name(user, project_name)
    if data:
        return data["services"]
    raise Exception("Something went wrong.\n" + error)



import numpy as np
import math
import geopandas as gpd
from datetime import datetime, timedelta
import pandas as pd

from scipy.ndimage import median_filter
from skimage.exposure import equalize_hist

from ..data.satellite import download_satellite_image
from ..data.satellite import explore_satellite_images
from ..processing import mask_raster
from ..processing import read_raster
from ..processing import colorize_raster
from ..processing import px_count
from ..processing import save_table
from ..processing import normalised_difference
from ..processing import convert_array_to_vector


def water_quality(image_name, aoi_mask, storage, desired_images=2, date=None):
    """
    This function calculates the water quality of a given image.

    It calculates the following layers:
        - Normalized Difference Water Index (NDWI) - ndwi_{date}.tif, ndwi_masked_{date}.tif, ndwi_categorized_{date}.tif
        - Normalized Difference Turbidity Index (NDTI) - ndti_{date}.tif, ndti_masked_{date}.tif, ndti_categorized_{date}.tif
        - Normalized Difference Chlorophyll Index (NDCI) - ndci_{date}.tif, ndci_masked_{date}.tif, ndci_categorized_{date}.tif
        - Median Water Mask - median_water_mask_{date}.tif
        - Dissolved Organic Carbon (DOC) - DOC_{date}.tif, DOC_masked_{date}.tif, DOC_categorized_{date}.tif

    It also calculates the following analytics:
        - Water Extent (in hectares and percentage) - table_water_extent.json
        - Water Turbidity (in hectares and percentage) - table_turbidity_Ha.json, table_turbidity_percent.json
        - Water Chlorophyll (in hectares and percentage) - table_chlorophyll_Ha.json, table_chlorophyll_percent.json
        - Water DOC (in hectares and percentage) - table_DOC_Ha.json, table_DOC_percent.json

    Parameters
    ----------
    image_name : str
        The name of the image to be processed.
    aoi_mask : dict
        The GeoJSON of the area of interest.
    storage : Storage
        The Storage object.
    desired_images : int, optional
        The number of images to be downloaded to determine the water body status, by default 2.
    date : str, optional
        The date of the image from which start downloading earlier images to determine the water body status,
        by default None, which uses the date given in the image name.

    Returns
    -------
    None

    """
    # read_raster image
    ds, raster = read_raster(image_name, storage, bands=[3, 4, 5, 8])
    date = image_name.split(".")[0].split("_")[-1]

    """ LAYERS """
    # NDWI

    # calculate ndwi
    ndwi = normalised_difference(raster, [1, 4])

    # save_raster ndwi
    raster_name_ndwi = f"ndwi_{date}.tif"
    storage.create(ndwi, raster_name_ndwi, ds=ds)

    # NDTI
    # calculate ndti
    ndti = normalised_difference(raster, [2, 1])

    # save_raster ndti
    raster_name_ndti = f"ndti_{date}.tif"
    storage.create(ndti, raster_name_ndti, ds=ds)

    # NDCI
    # calculate ndci
    ndci = normalised_difference(raster, [3, 2])
    ndci = equalize_hist(ndci)

    # save_raster ndci
    raster_name_ndci = f"ndci_{date}.tif"
    storage.create(ndci, raster_name_ndci, ds=ds)

    # MEDIAN WATER MASK
    # calculate median_water_mask
    median_water_mask, stored_images, _ = compute_median_water_mask(
        aoi_mask, desired_images, date, storage
    )  # aoi_mask is needed for downloading

    # save raster median_water_mask
    raster_name_median_water_mask = f"median_water_mask_{date}.tif"
    median_water_mask_path = storage.create(
        median_water_mask, raster_name_median_water_mask, ds=ds
    )

    # DOC
    # calculate doc
    doc_layers_array, doc_values_array, ds = compute_doc(stored_images, storage)

    # save_raster doc
    raster_name_doc = f"DOC_{date}.tif"
    storage.create(doc_layers_array[0], raster_name_doc, ds=ds)

    """ MASK LAYERS """
    shp = convert_array_to_vector(median_water_mask, median_water_mask_path)

    # mask_raster ndwi with aoi_mask
    ndwi_masked, _ = mask_raster(raster_name_ndwi, shp, storage)

    # save_raster ndwi_masked
    raster_name_ndwi_masked = f"ndwi_masked_{date}.tif"
    storage.create(ndwi_masked, raster_name_ndwi_masked, ds=ds)

    # mask_raster ndti with aoi_mask
    ndti_masked, _ = mask_raster(raster_name_ndti, shp, storage)
    ndti_masked[ndti_masked == False] = -9999  # convert false to nan

    # save_raster ndti_masked
    raster_name_ndti_masked = f"ndti_masked_{date}.tif"
    storage.create(ndti_masked, raster_name_ndti_masked, ds=ds)

    # mask_raster ndci with aoi_mask
    ndci_masked, _ = mask_raster(raster_name_ndci, shp, storage)
    ndci_masked[ndci_masked == False] = -9999  # convert false to nan

    # save_raster ndci_masked
    raster_name_ndci_masked = f"ndci_masked_{date}.tif"
    storage.create(ndci_masked, raster_name_ndci_masked, ds=ds)

    # mask_raster doc with aoi_mask
    doc_masked, _ = mask_raster(raster_name_doc, shp, storage)
    doc_masked[doc_masked == False] = -9999  # convert false to nan

    # save_raster doc_masked
    raster_name_doc_masked = f"DOC_masked_{date}.tif"
    storage.create(doc_masked, raster_name_doc_masked, ds=ds)

    """ ANALYTICS """
    # Water Extent
    extent_has = np.divide(median_water_mask.sum(axis=(1, 2)).max(), 100)
    update_extent_table(extent_has, date, storage)

    # Water Turbidity
    ndti_categories = np.digitize(ndti_masked, [-np.inf, -0.2, 0.4, np.inf])
    raster_name_ndti_categorized = f"ndti_categorized_{date}.tif"
    storage.create(ndti_categories, raster_name_ndti_categorized, ds=ds)

    ndti_colorized = colorize_raster(ndti_categories, colors=["green", "yellow", "red"])
    raster_name_ndti_colorized = f"ndti_categorized_rgb_{date}.tif"
    storage.create(ndti_colorized, raster_name_ndti_colorized, ds=ds)

    ndti_px_counted = px_count(ndti_categories, [1, 2, 3])
    ndti_has = np.divide(
        ndti_px_counted,
        100,
        out=np.zeros_like(ndti_px_counted, dtype=np.float64),
        where=100 != 0,
    )
    # save_table turbidity hectareas
    turbidity_table_name = "table_turbidity_Ha.json"
    turbidity_columns = ["Good [Has]", "Careful [Has]", "Bad [Has]", "Total [Has]"]
    save_table(
        data=ndti_has,
        columns=turbidity_columns,
        table_name=turbidity_table_name,
        date=date,
        storage=storage,
    )
    ndti_percent = [
        np.divide(ndti_has[i], ndti_has[-1]) * 100 for i in range(len(ndti_has) - 1)
    ]
    # save_table turbidity percent
    turbidity_percent_table_name = "table_turbidity_percent.json"
    turbidity_percent_columns = ["Good [%]", "Careful [%]", "Bad [%]"]
    save_table(
        data=ndti_percent,
        columns=turbidity_percent_columns,
        table_name=turbidity_percent_table_name,
        date=date,
        storage=storage,
    )

    # Water Chlorophyll
    ndci_categories = np.digitize(ndci_masked, [-np.inf, -0.3, 0.5, np.inf])
    raster_name_ndci_categorized = f"ndci_categorized_{date}.tif"
    storage.create(ndci_categories, raster_name_ndci_categorized, ds=ds)

    ndci_colorized = colorize_raster(ndci_categories, colors=["green", "yellow", "red"])
    raster_name_ndci_colorized = f"ndci_categorized_rgb_{date}.tif"
    storage.create(ndci_colorized, raster_name_ndci_colorized, ds=ds)

    ndci_px_counted = px_count(ndci_categories, [1, 2, 3])
    ndci_has = np.divide(
        ndci_px_counted,
        100,
        out=np.zeros_like(ndci_px_counted, dtype=np.float64),
        where=100 != 0,
    )
    # save_table chlorophyll hectareas
    chlorophyll_table_name = "table_chlorophyll_Ha.json"
    chlorophyll_columns = ["Good [Has]", "Careful [Has]", "Bad [Has]", "Total [Has]"]
    save_table(
        data=ndci_has,
        columns=chlorophyll_columns,
        table_name=chlorophyll_table_name,
        date=date,
        storage=storage,
    )
    ndci_percent = [
        np.divide(ndci_has[i], ndci_has[-1]) * 100 for i in range(len(ndci_has) - 1)
    ]
    # save_table chlorophyll percent
    chlorophyll_percent_table_name = "table_chlorophyll_percent.json"
    chlorophyll_percent_columns = ["Good [%]", "Careful [%]", "Bad [%]"]
    save_table(
        data=ndci_percent,
        columns=chlorophyll_percent_columns,
        table_name=chlorophyll_percent_table_name,
        date=date,
        storage=storage,
    )

    # DOC
    # Compute the mean of the doc values
    mean_mean_doc = np.mean([dict["mean_doc"] for dict in doc_values_array])
    mean_max_doc = np.mean([dict["max_doc"] for dict in doc_values_array])
    mean_min_doc = np.mean([dict["min_doc"] for dict in doc_values_array])
    mean_std_doc = np.mean([dict["std_doc"] for dict in doc_values_array])

    # Categorize the doc values
    a = 1
    min_to_N0 = -np.inf
    N0_to_N1 = mean_mean_doc + 2 * mean_std_doc / a
    N1_to_N2 = mean_mean_doc + 3 * mean_std_doc / a
    N2_to_N3 = mean_mean_doc + 4 * mean_std_doc / a
    N3_to_max = np.inf

    doc_categories = np.digitize(
        doc_masked, [min_to_N0, N0_to_N1, N1_to_N2, N2_to_N3, N3_to_max]
    )
    raster_name_doc_categorized = f"DOC_categorized_{date}.tif"
    storage.create(doc_categories, raster_name_doc_categorized, ds=ds)

    doc_colorized = colorize_raster(
        doc_categories, colors=["green", "yellow", "orange", "red"]
    )
    raster_name_doc_colorized = f"DOC_categorized_rgb_{date}.tif"
    storage.create(doc_colorized, raster_name_doc_colorized, ds=ds)

    doc_px_counted = px_count(doc_categories, [1, 2, 3, 4])
    doc_has = np.divide(
        doc_px_counted,
        100,
        out=np.zeros_like(doc_px_counted, dtype=np.float64),
        where=100 != 0,
    )
    # save_table DOC hectareas
    doc_table_name = "table_DOC_Ha.json"
    doc_columns = ["N0 [Has]", "N1 [Has]", "N2 [Has]", "N3 [Has]", "Total [Has]"]
    save_table(
        data=doc_has,
        columns=doc_columns,
        table_name=doc_table_name,
        date=date,
        storage=storage,
    )
    doc_percent = [
        np.divide(doc_has[i], doc_has[-1]) * 100 for i in range(len(doc_has) - 1)
    ]
    # save_table DOC percent
    doc_percent_table_name = "table_DOC_percent.json"
    doc_percent_columns = ["N0 [%]", "N1 [%]", "N2 [%]", "N3 [%]"]
    save_table(
        data=doc_percent,
        columns=doc_percent_columns,
        table_name=doc_percent_table_name,
        date=date,
        storage=storage,
    )


"""
Methods for Layers: Median Water Mask
"""


def compute_median_water_mask(aoi_mask, desired_images, from_date_to_earlier, storage):
    # Get the bbox from the aoi_mask
    bbox = gpd.GeoDataFrame.from_features(aoi_mask["features"]).geometry[0]

    stored_images = download_missing_images(
        desired_images=desired_images,
        from_date_to_earlier=from_date_to_earlier,
        bbox=bbox,
        storage=storage,
    )
    water_masks_array, ds = compute_water_masks(stored_images, storage)
    median_water_mask = median_filter(water_masks_array, size=(7, 3, 3))
    binary_median_water_mask = median_water_mask.astype(np.uint8)

    # save_raster
    median_raster_name = "median_water_mask.tif"
    storage.create(median_water_mask, median_raster_name, ds=ds)
    return binary_median_water_mask, stored_images, water_masks_array


def download_missing_images(desired_images, from_date_to_earlier, bbox, storage):
    # Convert the date string to a datetime object
    date_obj = datetime.strptime(from_date_to_earlier, "%Y-%m-%d")
    # Check how many images that begin with "S2L2A" are in storage
    files = storage.list()
    stored_images = []
    for file in files:
        # If the file is a Sentinel-2 image of a date equal or earlier than 'from_date_to_earlier' , add it to the list:
        if file.startswith("S2L2A"):
            image_date = datetime.strptime(file.split("_")[1].split(".")[0], "%Y-%m-%d")
            if image_date <= date_obj and len(stored_images) < desired_images:
                stored_images.append(file)
    # If images are less than desired, download the missing ones
    images_found = []
    while len(stored_images) < desired_images:
        time_interval = (
            (date_obj - timedelta(days=10)).strftime("%Y-%m-%d"),
            date_obj.strftime("%Y-%m-%d"),
        )
        images_found = explore_satellite_images(
            bbox, time_interval=time_interval, cloud_cover=5
        )
        for image in images_found:
            if f"S2L2A_{image['date'].split('T')[0]}.tif" not in stored_images:
                path = download_satellite_image(
                    storage, bbox, image["date"].split("T")[0]
                )
                stored_images.append(path)
        date_obj -= timedelta(days=10)
    return stored_images


def compute_water_masks(stored_images, storage):
    water_masks_array = []
    for path in stored_images:
        image_name = path.split("/")[-1]
        date = image_name.split("_")[1].split(".")[0]
        # read raster
        ds, raster = read_raster(image_name, storage, bands=[3, 4, 5, 8])

        # calculate ndwi
        ndwi = normalised_difference(raster, [1, 4])

        # apply thershold to calculate water-mask
        threshold = 0
        water_mask = ndwi >= threshold

        # save_raster water_mask
        raster_name_water_mask = f"water_mask_{date}.tif"
        storage.create(water_mask, raster_name_water_mask, ds=ds)
        water_masks_array.append(water_mask)

    return water_masks_array, ds


""" 
Methods for Layers: DOC
"""


def compute_doc(stored_images, storage):
    doc_layers_array = []
    doc_values_array = []
    for path in stored_images:
        image_name = path.split("/")[-1]
        # read raster
        ds, raster = read_raster(image_name, storage, bands=[3, 4, 5, 8])

        # calculate doc for each image
        # calculate doc
        bands = [1, 2]
        bands = np.array(bands) - 1
        # Separate the bands
        band1 = raster[bands[0], :, :]
        band2 = raster[bands[1], :, :]
        # convert the bands to floats
        band1 = band1.astype(float)
        band2 = band2.astype(float)
        doc = 432 * pow(math.e, -2.24 * band1 / band2)
        mean_doc = np.nanmean(doc)
        max_doc = np.nanmax(doc)
        min_doc = np.nanmin(doc)
        std_doc = np.nanstd(doc)
        doc_values_dict = {
            "mean_doc": mean_doc,
            "max_doc": max_doc,
            "min_doc": min_doc,
            "std_doc": std_doc,
        }
        doc_values_array.append(doc_values_dict)
        doc_layers_array.append(doc)
    return doc_layers_array, doc_values_array, ds


"""
Methods for Analytics: Water Extent
"""


def update_df(df):
    df["Total [Has]"] = df["Water [Has]"].max()
    df["Not Water [Has]"] = df["Total [Has]"] - df["Water [Has]"]
    df["Percentage [%]"] = df["Water [Has]"] / df["Total [Has]"] * 100
    return df


def update_extent_table(extent_has, date, storage):
    if f"table_water_extent.json" not in storage.list():
        df = pd.DataFrame(
            {
                "Water [Has]": extent_has,
                "Not Water [Has]": 0,
                "Total [Has]": extent_has,
                "Percentage [%]": 100,
            },
            index=[date],
        )
        df = update_df(df)
    else:
        df = storage.read(f"table_water_extent.json")
        if date in df.index:
            df.loc[date, "Water [Has]"] = extent_has
            update_df(df)
            return df
        if isinstance(df.index, pd.DatetimeIndex):
            # Handle pd.DateTimeIndex, converting index to string
            df.index = df.index.strftime("%Y-%m-%d")
        new_row = pd.DataFrame(
            {
                "Water [Has]": extent_has,
                "Not Water [Has]": 0,
                "Total [Has]": 0,
                "Percentage [%]": 0,
            },
            index=[date],
        )
        df = pd.concat([new_row, df.loc[:]])
        df = update_df(df)
    storage.create(df, f"table_water_extent.json")
    return df



import streamlit as st
import pandas as pd
import requests
import pydeck as pdk
import os

# tengo que poder sacar la url de alguna manera

# API_URL = "http://localhost"
# # API_URL = "http://148.113.137.45"
# ANALYTICS_URL = f"{API_URL}:8011"
# XYZ_URL = f"{API_URL}:8012"

st.set_page_config(page_title="SPAI Demo", page_icon="🌍")


@st.cache_data(ttl=10)
def get_data():  # in cloud fails because localhost is inside docker, need public url
    api_url = os.getenv("ANALYTICS_URL")
    analytics = requests.get(api_url).json()
    df = pd.DataFrame(analytics)
    return df


df = get_data()

# AWS Open Data Terrain Tiles
TERRAIN_IMAGE = (
    "https://s3.amazonaws.com/elevation-tiles-prod/terrarium/{z}/{x}/{y}.png"
)

# Define how to parse elevation tiles
ELEVATION_DECODER = {"rScaler": 256, "gScaler": 1, "bScaler": 1 / 256, "offset": -32768}

st.sidebar.markdown("### Dates")

selected_layers = [
    pdk.Layer(
        "TerrainLayer",
        texture=f"{os.getenv('XYZ_URL')}/NDVI_{date}.tif/{{z}}/{{x}}/{{y}}.png",
        elevation_decoder=ELEVATION_DECODER,
        elevation_data=TERRAIN_IMAGE,
    )
    for date in df.index
    if st.sidebar.checkbox(date, True)
]

if selected_layers:
    st.pydeck_chart(
        pdk.Deck(
            map_style="mapbox://styles/mapbox/light-v9",
            initial_view_state={
                "latitude": 41.4,
                "longitude": 2.17,
                "zoom": 9,
                "pitch": 60,
            },
            layers=selected_layers,
        )
    )
else:
    st.error("Please choose at least one layer above.")

st.title("Vegetation Analytics")

st.line_chart(df)
if st.checkbox("Show data"):
    st.write(df)



# Updated read_raster function to handle invalid extensions
def read_raster(image_name, storage, bands=[8, 4]):
    ds = storage.read(image_name)
    if ds is None:
        raise TypeError("Not a valid raster file")
    raster = ds.read(bands)
    return ds, raster



"""
NASA Fire Weather Index
"""
from datetime import datetime, timedelta
from typing import List, Union

import geopandas as gpd
import pandas as pd
import xarray as xr

URL = "https://portal.nccs.nasa.gov/datashare/GlobalFWI/v2.0/fwiCalcs.GEOS-5/Default/GPM.LATE.v5"

# TODO: Add attr to the FWI


def curate_nasa_fwi(ds: xr.Dataset, bounds: List, date: datetime) -> xr.DataArray:
    """
    Curate the NASA FWI data, clipping by the bounds and setting the crs

    Parameters
    ----------
    ds : xarray.Dataset
        The dataset containing the FWI data
    bounds : list
        The bounds of the area of interest
    date : datetime
        The date of the data

    Returns
    -------
    xarray.DataArray
        The curated data array
    """
    fwi = ds["GPM.LATE.v5_FWI"].sel(time=1)
    # Add or remove 0.1 to the bounds to make sure we get all the data
    fwi = fwi.sel(
        lon=slice(bounds[0] - 0.1, bounds[2] + 0.1),
        lat=slice(bounds[1] - 0.1, bounds[3] + 0.1),
    )
    fwi.rio.set_spatial_dims(x_dim="lon", y_dim="lat", inplace=True)
    fwi.rio.write_crs("epsg:4326", inplace=True)
    fwi.coords["time"] = pd.to_datetime(date)

    return fwi


def get_nasa_fwi(dates: List, gdf: Union[str, gpd.GeoDataFrame]) -> xr.Dataset:
    """
    Get the NASA FWI data for the given dates and area of interest

    Parameters
    ----------
    dates : list
        The start and end dates of the data
    aoi : str
        The path to the area of interest file

    Returns
    -------
    xarray.Dataset
        The dataset containing the FWI data
    """
    dates = [datetime.strptime(date, "%Y-%m-%d") for date in dates]

    if isinstance(gdf, str):
        gdf = gpd.read_file(gdf)
    bounds = gdf.total_bounds

    current_date = dates[0]
    data_arrays = []
    while current_date <= dates[1]:
        year = current_date.year
        date_formatted = datetime.strftime(current_date, "%Y%m%d")
        file_url = (
            f"{URL}/{year}/FWI.GPM.LATE.v5.Daily.Default.{date_formatted}.nc#mode=bytes"
        )
        try:
            ds = xr.open_dataset(file_url)
        except OSError:
            current_date += timedelta(days=1)
            continue
        try:
            fwi_data_array = curate_nasa_fwi(ds, bounds, current_date)
        except KeyError:
            current_date += timedelta(days=1)
            continue
        data_arrays.append(fwi_data_array)
        current_date += timedelta(days=1)

    try:
        combined = xr.concat(data_arrays, dim="time").to_dataset(name="fwi")
    except ValueError:
        return None

    combined.rio.set_spatial_dims(x_dim="lon", y_dim="lat", inplace=True)
    combined.rio.write_crs("epsg:4326", inplace=True)

    return combined


def download_nasa_fwi(dates: List, aoi: str, storage):
    """
    Download the NASA FWI data for the given dates and area of interest

    Parameters
    ----------
    dates : list
        The start and end dates of the data
    aoi : str
        The path to the area of interest file
    storage : BaseStorage
        The storage object
    """
    fwi = get_nasa_fwi(dates, aoi)
    curated_fwi = fwi["fwi"].to_dataset(name="fwi")
    storage.create_from_zarr(curated_fwi, name="fwi")



from ..repos import APIRepo
from .get_project_by_name import get_project_by_name
from .get_service_by_name_type_project import get_service_by_name_type_project


def get_logs(user, project_name, service_type, service_name):
    project = get_project_by_name(user, project_name)
    service = get_service_by_name_type_project(
        user, project["id"], service_type, service_name
    )
    repo = APIRepo()
    return repo.get_logs(user, service["id"])



import shutil
from pathlib import Path
import json


def init_project(path, project_name):
    # copy template
    template = Path(__file__).parent / "project-template"
    shutil.copytree(template, path / project_name)
    # change name to project in spai.config.yaml
    config = path / project_name / "spai.config.yaml"
    # read json and change name
    with open(config, "r") as f:
        json = f.read()
    json = json.replace("project-template", project_name)
    with open(config, "w") as f:
        f.write(json)



import matplotlib
import numpy as np


def colorize_raster(raster, colors=["darkgreen"]):
    raster = raster.squeeze(axis=0)

    # Define color mapping
    color_mapping = {i: colors[i - 1] for i in range(1, len(colors) + 1)}

    # Create an array for colors
    colored_image = np.zeros((raster.shape[0], raster.shape[1], 4))
    for value, color in color_mapping.items():
        colored_image[raster == value] = matplotlib.colors.to_rgba(color)

    return colored_image



import rasterio as rio
from rasterio.features import shapes
import numpy as np
import geopandas as gpd


def convert_array_to_vector(mask: np.array, img_path: str) -> gpd.GeoDataFrame:
    """Convert masked array to vector file(.shp).

    Args:
        binary_mask: Output predicted binary mask.
        img_path: Path of the raster file used for the prediction.

    Returns:
        shp: GeoPandas DataFrame.

    """
    with rio.open(img_path, mode='r') as src:

        # create a dict of features generator
        features = (
            {'properties': {'value': value}, 'geometry': shape}
            for (shape, value) in shapes(
                mask.astype(np.uint8), mask=mask, transform=src.transform
            )
        )

        shp = gpd.GeoDataFrame.from_features(
            features,
            crs=src.crs,  # no need to convert to string
            columns=['geometry', 'value']
        )

    return shp


import pandas as pd


def save_table(data, columns, table_name, date, storage):
    # columns number must be equal to data number of columns
    if len(columns) != len(data):
        raise ValueError("columns number must be equal to data number of columns")
    # check if table exists in storage
    if table_name in storage.list():
        # append data to table
        base_df = storage.read(table_name)
        if isinstance(base_df.index, pd.DatetimeIndex):
            # Handle pd.DateTimeIndex, converting index to string
            base_df.index = base_df.index.strftime("%Y-%m-%d")
        df = pd.concat(
            [pd.DataFrame([data], columns=columns, index=[date]), base_df.loc[:]]
        )
    else:
        # create table with data
        df = pd.DataFrame([data], columns=columns, index=[date])
    storage.create(data=df, name=table_name)
    return df



from ..repos import APIRepo


def get_service_by_name_type_project(user, project_id, service_type, service_name):
    repo = APIRepo()
    data, error = repo.retrieve_service_by_name_type_project(
        user, project_id, service_type, service_name
    )
    if data:
        return data
    raise Exception("Something went wrong.\n" + error)



import os
from datetime import timedelta
from sentinelhub import (
    SHConfig,
    MimeType,
    CRS,
    BBox,
    SentinelHubRequest,
    bbox_to_dimensions,
    BBoxSplitter,
)
import glob
from datetime import datetime
import math


class SHDownloader:
    def __init__(self, download_folder):
        self.download_folder = download_folder
        self.config = SHConfig()
        self.config.sh_client_id = os.environ["SH_CLIENT_ID"]
        self.config.sh_client_secret = os.environ["SH_CLIENT_SECRET"]
        self.mosaicking_order = None

    def compute_image_size(self, gdf):
        self.bbox = BBox(bbox=gdf.total_bounds.tolist(), crs=CRS.WGS84)
        self.bbox_size = bbox_to_dimensions(self.bbox, resolution=self.resolution)
        return self.bbox_size

    def prepare_time_interval(self, date):
        if not date:
            # Assuming is DEM data
            self.time_interval = ('2011-01-01', '2015-01-07')
            return
        date = datetime.strptime(date, "%Y-%m-%d")
        date_day_before = date - timedelta(days=1)
        date_next_day = date + timedelta(days=1)
        date_day_before = date_day_before.strftime("%Y-%m-%d")
        date_next_day = date_next_day.strftime("%Y-%m-%d")
        self.time_interval = (date_day_before, date_next_day)

    def download(self, gdf, date):
        self.prepare_time_interval(date)
        self.compute_image_size(gdf)
        if self.bbox_size[0] > 2500 or self.bbox_size[1] > 2500:
            # return self.download_large_area(gdf)
            raise Exception("Area too large")
        return self.download_small_area(self.bbox)

    def request_bands(self, bbox):
        return SentinelHubRequest(
            data_folder=self.download_folder,
            evalscript=self.script,
            input_data=[
                SentinelHubRequest.input_data(
                    data_collection=self.data_collection,
                    time_interval=self.time_interval,
                    mosaicking_order=self.mosaicking_order,
                )
            ],
            responses=[SentinelHubRequest.output_response("default", MimeType.TIFF)],
            bbox=self.bbox,
            size=bbox_to_dimensions(bbox, resolution=self.resolution),
            config=self.config,
        )

    def download_small_area(self, bbox):
        request_bands = self.request_bands(bbox)
        request_bands.save_data()
        downloaded_files = glob.glob(f"{self.download_folder}/*/response.tiff")
        assert len(downloaded_files) > 0, "No files downloaded"
        # assert len(downloaded_files) == 1, "There should be only one response file"
        folder = downloaded_files[0]
        return folder

    def get_tiles(self):
        # compute number of images to download
        self.max_resolution = 2500
        return (
            math.ceil(self.bbox_size[0] / self.max_resolution),
            math.ceil(self.bbox_size[1] / self.max_resolution),
        )

    def download_large_area(self, gdf):
        tiles = self.get_tiles()
        geom = gdf.geometry.unary_union  # ???
        bbox_splitter = BBoxSplitter([geom], CRS.WGS84, tiles)
        bbox_list = bbox_splitter.get_bbox_list()
        for i, bbox in enumerate(bbox_list):
            request_bands = self.request_bands(bbox)
            request_bands.save_data()
        downloaded_files = glob.glob(f"{self.download_folder}/*/response.tiff")
        dst_path = self.download_folder + "/output.tiff"
        command = f"gdal_merge.py -o {dst_path} " + " ".join(downloaded_files)
        os.system(command)
        return dst_path



from pathlib import Path
import os
import json
import jwt


class AuthRepo:
    def __init__(self):
        self.algorithms = ["RS256"]
        self.home = str(Path.home())
        self.creds_path = self.home + "/.cache/spai/creds.json"

    def save_creds(self, data):
        os.makedirs(self.home + "/.cache/spai", exist_ok=True)
        with open(self.creds_path, "w") as f:
            json.dump(data, f)
        return self.creds_path

    def load_creds(self):
        if os.path.exists(self.creds_path):
            with open(self.creds_path, "r") as f:
                creds = json.load(f)
            user = self.decode_token(creds)
            user["id_token"] = creds["id_token"]
            return user
        return None

    def decode_token(self, token_data):
        return jwt.decode(
            token_data["id_token"],
            algorithms=self.algorithms,
            options={"verify_signature": False},
        )

    def logout(self):
        os.remove(self.creds_path)



import rasterio


def mask_raster(raster_name, gdf, storage):
    # crop image to geometry
    ds = storage.read(raster_name)
    if ds is None:
        raise Exception("Raster not found")
    geometry = gdf.geometry
    return rasterio.mask.mask(ds, geometry)  # importar así para testing



from .init_project import init_project
from .install_requirements import install_requirements
from .run_local import run_local
from .deploy_folder import deploy_folder
from .deploy_template import deploy_template
from .get_services import get_services
from .get_projects import get_projects
from .stop_service import stop_service, stop_service_by_name
from .delete_project import delete_project
from .get_logs import get_logs
from .download_template import download_template



MIMETYPES = {
    ".jpg": "image/jpeg",
    ".jpeg": "image/jpeg",
    ".png": "image/png",
    ".gif": "image/gif",
    ".webp": "image/webp",
    ".html": "text/html",
    ".htm": "text/html",
    ".css": "text/css",
    ".js": "application/javascript",
    ".geojson": "application/geo+json",
    ".json": "application/json",
    ".xml": "application/xml",
    ".pdf": "application/pdf",
    ".txt": "text/plain",
    ".csv": "text/csv",
    ".mp4": "video/mp4",
    ".mp3": "audio/mpeg",
    ".wav": "audio/wav",
    ".zip": "application/zip",
    ".tar": "application/x-tar",
    ".doc": "application/msword",
    ".docx": "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
    ".xls": "application/vnd.ms-excel",
    ".xlsx": "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
    ".ppt": "application/vnd.ms-powerpoint",
    ".pptx": "application/vnd.openxmlformats-officedocument.presentationml.presentation",
    ".mdb": "application/vnd.ms-access",
    ".pub": "application/x-mspublisher",
    ".vsd": "application/vnd.visio",
    ".msg": "application/vnd.ms-outlook",
    ".rtf": "application/rtf",
    ".odt": "application/vnd.oasis.opendocument.text",
    ".ods": "application/vnd.oasis.opendocument.spreadsheet",
    ".odp": "application/vnd.oasis.opendocument.presentation",
}



from typing import Union, Optional

from .sentinelhub import (
    SHS2L2ADownloader,
    SHS2L1CDownloader,
    SHS1Downloader,
    SHDEM30Downloader,
    SHCloudMaskDownloader,
)
from .utils import create_aoi_geodataframe
import shutil
import os
import geopandas as gpd
from .explore import explore_satellite_images
from ...storage import Storage


def download_satellite_image(storage, aoi, date=None, sensor="S2L2A", **kwargs):
    if date is None:
        options = explore_satellite_images(aoi, sensor=sensor, **kwargs)
        if len(options) == 0:
            raise ValueError("No images found")
        date = options[0]["date"].split("T")[0]
    gdf = create_aoi_geodataframe(aoi)
    if isinstance(gdf, gpd.GeoDataFrame):
        download_dir = "/tmp/sentinelhub"
        if sensor == "S2L2A":
            downloader = SHS2L2ADownloader(download_dir)
        elif sensor == "S2L1C":
            downloader = SHS2L1CDownloader(download_dir)
        elif sensor == "S1":
            downloader = SHS1Downloader(download_dir)
        else:
            raise Exception(f"sensor {sensor} not supported")
        if isinstance(date, str):
            dst_path = downloader.download(gdf, date)
            dst_path = storage.create(dst_path, name=f"{sensor}_{date}.tif")
            shutil.rmtree(download_dir)
            return dst_path
        # elif isinstance(date,list):
        #     dst_paths = []
        #     for d in date:
        #         dst_path = downloader.download(gdf, d)
        #         dst_path = storage.create(dst_path, name=f"{sensor}_{d}.tif")
        #         dst_paths.append(dst_path)
        #     shutil.rmtree(download_dir)
        #     return dst_paths
        else:
            raise Exception(f"Date type {type(date)} not supported")
    else:
        return gdf


def download_dem(
    storage: Storage, aoi: Union[str, gpd.GeoDataFrame], **kwargs
) -> Union[str, gpd.GeoDataFrame]:
    """
    Download the DEM from a given area of interest

    Parameters
    ----------
    storage : Storage
        Storage object
    aoi : Union[str, gpd.GeoDataFrame]
        Area of interest
    kwargs
        Additional arguments

    Returns
    -------
    Union[str, gpd.GeoDataFrame]
        Path to the DEM file or the GeoDataFrame of the area of interest
    """
    gdf = create_aoi_geodataframe(aoi)
    if isinstance(gdf, gpd.GeoDataFrame):
        download_dir = "/tmp/sentinelhub"
        downloader = SHDEM30Downloader()
        dst_path = downloader.download(gdf, date=None)
        dst_path = storage.create(dst_path, name="DEM.tif")
        shutil.rmtree(download_dir) if os.path.exists(download_dir) else None
        return dst_path
    else:
        return gdf


def download_cloud_mask(
    storage: Storage,
    aoi: Union[str, gpd.GeoDataFrame],
    date: Optional[str] = None,
    **kwargs,
) -> Union[str, gpd.GeoDataFrame]:
    """
    Download the cloud mask from a given area of interest

    Parameters
    ----------
    storage : Storage
        Storage object
    aoi : Union[str, gpd.GeoDataFrame]
        Area of interest
    date : Optional[str]
        Date of the image
    kwargs
        Additional arguments

    Returns
    -------
    Union[str, gpd.GeoDataFrame]
        Path to the cloud mask file or the GeoDataFrame of the area of interest
    """
    if date is None:
        options = explore_satellite_images(aoi, sensor=sensor, **kwargs)
        if len(options) == 0:
            raise ValueError("No images found")
        date = options[0]["date"].split("T")[0]
    gdf = create_aoi_geodataframe(aoi)
    if isinstance(gdf, gpd.GeoDataFrame):
        download_dir = "/tmp/sentinelhub"
        downloader = SHCloudMaskDownloader(download_dir)
        dst_path = downloader.download(gdf, date)
        dst_path = storage.create(dst_path, name=f"cloud_mask_{date}.tif")
        shutil.rmtree(download_dir)
        return dst_path
    else:
        return gdf



from ..repos import APIRepo


def stop_service(user, project, service_id):
    repo = APIRepo()
    service = repo.retrieve_service(user, service_id)
    print(f"Stopping service {service['type']}/{service['name']} ...")
    return repo.stop_service(user, service_id)


def stop_service_by_name(user, project, service_type, service_name):
    repo = APIRepo()
    data, error = repo.retrieve_project_by_name(user, project)
    print(f"Stopping service {service_type}/{service_name} ...")
    data, error = repo.stop_service_by_name(
        user, data["id"], service_type, service_name
    )
    if error:
        raise Exception("Something went wrong.\n" + error)
    return data



from fastapi.middleware.cors import CORSMiddleware
from fastapi import FastAPI
from starlette.responses import StreamingResponse
from fastapi import HTTPException, status
import uvicorn
import argparse

from spai.image.xyz import get_image_data, get_tile_data, ready_image
from spai.image.xyz.errors import ImageOutOfBounds
from spai.storage import Storage

# init api
app = FastAPI(title="xyz")

# configure CORS
origins = [
    "*",
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

storage = Storage("data")


@app.get("/")
def retrieve_images():
    return storage.list("*.tif")


@app.get("/{image}/{z}/{x}/{y}.png")
def retrieve_image_tile(
    image: str,
    z: int,
    x: int,
    y: int,
    bands: str = "1",
    stretch: str = "0,1",
    palette: str = "viridis",
):
    image_path = storage.get_path(image)
    tile_size = (256, 256)
    if len(bands) == 1:
        bands = int(bands)
    else:
        bands = tuple([int(band) for band in bands.split(",")])
    stretch = tuple([float(v) for v in stretch.split(",")])
    try:
        tile = get_tile_data(image_path, (x, y, z), bands, tile_size)
        tile = get_image_data(tile, stretch, palette)
        image = ready_image(tile)
        return StreamingResponse(image, media_type="image/png")
    except ImageOutOfBounds as error:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=error.message)


# need this to run in background
if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--host", type=str, default="0.0.0.0")
    parser.add_argument("--port", type=int, default=8000)
    args = parser.parse_args()
    uvicorn.run(app, host=args.host, port=args.port)



from .SHExplorer import SHExplorer
from .SHS2L2ADownloader import SHS2L2ADownloader
from .SHS2L1CDownloader import SHS2L1CDownloader
from .SHS1Downloader import SHS1Downloader
from .SHDEM30Downloader import SHDEM30Downloader
from .SHCloudMaskDownloader import SHCloudMaskDownloader



import requests
import zipfile
import io
import shutil

from ..repos import APIRepo


def download_template(user, template, template_path, force):
    repo = APIRepo()
    response = repo.retrieve_template(user, template)
    if response.status_code == 200:
        data = response.json()
        # raise error if already exists
        path = template_path / data["name"]
        if path.exists():
            print("Project already exists, use -f to re-download")
            if force:
                shutil.rmtree(path)
        if not path.exists() or force:
            # clone github repo with template
            print(f"Downloading {data['name']}...")
            zip_url = (
                data["url"] + "/archive/main.zip"
            )  # OJO asume que la rama se llama main !!!
            r = requests.get(zip_url, stream=True)
            z = zipfile.ZipFile(io.BytesIO(r.content))
            z.extractall(path=template_path)
            # move files to template_path
            repo_name = data["url"].split("/")[-1]
            branch = "main"  # TODO: cambiar
            shutil.move(f"{repo_name}-{branch}", path)
        return path
    raise Exception(response.json()["detail"])



class LoginError(Exception):
    def __init__(self, message="Login Error"):
        self.message = message
        super().__init__(self.message)


class AuthTimeOut(Exception):
    def __init__(self, message="Authentication timed out"):
        self.message = message
        super().__init__(self.message)



from .SHDownloader import SHDownloader
from sentinelhub import DataCollection


class SHS1Downloader(SHDownloader):

    def __init__(self, download_folder='/tmp'):
        super().__init__(download_folder)
        self.data_collection = DataCollection.SENTINEL1
        self.resolution = 10  # mpp
        self.script = """
            //VERSION=3
            function setup() {
                return {
                    input: [{
                        bands: ["VV","VH"],
                        units: "LINEAR_POWER"
                    }],
                    output: {
                        bands: 2,
                        sampleType: "INT16"
                    }
                };
            }

            function evaluatePixel(sample) {
                return [sample.VV,
                        sample.VH];
            }
        """


import typer
from pathlib import Path

# Add the cli directory to the Python path
# spai_cli_dir = os.path.dirname(os.path.realpath(__file__))
# sys.path.append(os.path.join(spai_cli_dir))

from .commands import auth as _auth
from .commands import list
from .repos import APIRepo
from .project import (
    init_project,
    install_requirements,
    run_local,
    deploy_folder,
    get_logs,
    download_template,
    deploy_template,
    get_services,
    stop_service,
    stop_service_by_name,
    delete_project,
)
from .auth import auth
from . import __version__

app = typer.Typer(help="Welcome to SPAI")
app.add_typer(_auth.app, name="auth")
app.add_typer(list.app, name="list")


@app.command(help="Create a new project from starter template")
def init(path: Path = typer.Option(Path.cwd(), "-p", "--path", help="Project path")):
    try:
        project_name = typer.prompt("Project name")
        init_project(path, project_name)
        typer.echo(f"Project {project_name} created at {path}")
    except Exception as e:
        typer.echo(f"Error: {e}")


@app.command(help="Install requirements for a project")
def install(
    path: Path = typer.Argument(Path.cwd(), help="Project path"),
    verbose: bool = typer.Option(False, "-v", "--verbose"),
):
    try:
        install_requirements(path, typer, verbose)
    except Exception as e:
        typer.echo(f"Error: {e}")


@app.command(help="Run a project locally")
def run(
    path: Path = typer.Argument(Path.cwd(), help="Project path"),
    template: str = typer.Option(None, "-t", "--template", help="Template name"),
    template_path: Path = typer.Option(
        Path.cwd(),
        "-p",
        "--path",
        help="Destination path for the project created from the template",
    ),
    force: bool = typer.Option(False, "-f", "--force", help="Force download template"),
):
    try:
        if template is not None:
            user = auth()
            path = download_template(user, template, template_path, force)
        return run_local(path, typer)
    except Exception as e:
        typer.echo(f"Error: {e}")


@app.command(help="Deploy a project to the cloud")
def deploy(
    path: Path = typer.Argument(Path.cwd(), help="Project path"),
    template: str = typer.Option(None, "-t", "--template", help="Template name"),
    verbose: bool = typer.Option(False, "-v", "--verbose"),
):
    try:
        user = auth()
        if template:
            return deploy_template(user, template)
        return deploy_folder(user, path, typer, verbose)
    except Exception as e:
        typer.echo(f"Error: {e}")


@app.command(help="Retrieve the logs of a service")
def logs(
    project: str = typer.Argument(..., help="Project name"),
    service_type: str = typer.Argument(
        ..., help="Service type (script, api, ui, etc.)"
    ),
    service_name: str = typer.Argument(..., help="Service name"),
):
    try:
        user = auth()
        logs = get_logs(user, project, service_type, service_name)
        typer.echo(logs)
    except Exception as e:
        typer.echo(f"Error: {e}")


@app.command(help="Clone a template")
def clone(
    template: str = typer.Argument(..., help="Template name"),
    path: Path = typer.Option(
        Path.cwd(),
        "-p",
        "--path",
        help="Destination path for the project created from the template",
    ),
    force: bool = typer.Option(False, "-f", "--force", help="Force download template"),
):
    try:
        user = auth()
        path = download_template(user, template, path, force)
        typer.echo(f"Template available at {path}.")
    except Exception as e:
        typer.echo(f"Error: {e}")


@app.command(help="Stop a project")
def stop(
    project: str = typer.Argument(..., help="Project name"),
    service_type: str = typer.Argument(None, help="Service type"),
    service_name: str = typer.Argument(None, help="Service name"),
):
    try:
        user = auth()
        if project is None:
            raise Exception("Project name is required.")
        services = get_services(user, project)
        if len(services) == 0:
            return typer.echo(f"No services running in project '{project}'.")
        if service_type and service_name:
            return stop_service_by_name(user, project, service_type, service_name)
        typer.echo(f"Stopping all services in project '{project}'...")
        for service_id in services:
            stop_service(user, project, service_id)
        return typer.echo(f"Stopped all services in project '{project}'.")
    except Exception as e:
        return typer.echo(e)


@app.command(help="Delete a project")
def delete(
    project: str = typer.Argument(..., help="Project name"),
):
    try:
        user = auth()
        if project is None:
            raise Exception("Project name is required.")
        delete_project(user, project)
        return typer.echo(f"Deleted project '{project}'.")
    except Exception as e:
        return typer.echo(e)


@app.command(help="Get SPAI version")
def version():
    typer.echo(f"SPAI Version: {__version__}")


@app.command(help="Get SPAI API url and status")
def api():
    repo = APIRepo()
    typer.echo(f"SPAI API URL: {repo.url}")
    typer.echo(repo.info())


if __name__ == "__main__":
    app()



import typer
from ..project import get_projects, get_services
from ..auth import auth


app = typer.Typer(help="Retrive information about your projects and services")


@app.command(help="List all projects")
def projects():
    try:
        user = auth()
        projects = get_projects(user)
        if len(projects) == 0:
            return typer.echo(f"No projects found.")
        typer.echo("Projects:")
        for project in projects:
            typer.echo(project)
    except Exception as e:
        return typer.echo(e)


@app.command(help="List all services in a project")
def services(
    project: str = typer.Argument(..., help="Project name"),
):
    try:
        user = auth()
        if project is None:
            raise Exception("Project name is required.")
        services = get_services(user, project)
        if len(services) == 0:
            return typer.echo(f"No services running in project '{project}'.")
        return typer.echo(f"Services in project '{project}': {services}")
    except Exception as e:
        return typer.echo(e)


if __name__ == "__main__":
    app()



from pydantic import BaseModel
from typing import Union, List
from pathlib import Path
from typing import Optional

from .StorageConfig import StorageConfig


class ScriptConfig(BaseModel):
    name: str
    run_on_start: bool = True
    command: Optional[str] = None
    run_every: Optional[int] = None  # seconds (in cloud minutes)
    storage: Optional[str] = None  # folder to bind in cloud
    type: str = "script"


class NotebookConfig(BaseModel):
    name: str
    command: Union[str, None] = None
    run_every: Union[int, None] = None  # seconds
    run_on_start: bool = True
    storage: Union[str, None] = None  # folder to bind in cloud
    type: str = "notebook"


class APIConfig(BaseModel):
    name: str
    command: Union[str, None] = None
    port: int = 8000
    host: str = "0.0.0.0"
    storage: Union[str, None] = None  # folder to bind in cloud
    type: str = "api"


class UIConfig(BaseModel):
    name: str
    command: str  # steamlit, javascript, ...
    port: int = 3000
    env: dict = {}  # can accept the name of another service as a url placeholder
    type: str = "ui"


class Config(BaseModel):
    dir: Path
    project: str
    scripts: List[ScriptConfig] = []
    notebooks: List[NotebookConfig] = []
    apis: List[APIConfig] = []
    uis: List[UIConfig] = []
    storage: List[StorageConfig] = []

    # iterator for all the services
    def __iter__(self):
        if self.scripts:
            for script in self.scripts:
                yield script
        if self.notebooks:
            for notebook in self.notebooks:
                yield notebook
        if self.apis:
            for api in self.apis:
                yield api
        if self.uis:
            for ui in self.uis:
                yield ui

    def type2folder(self, type):
        return type + "s"



from .SHDownloader import SHDownloader
from sentinelhub import MosaickingOrder, DataCollection


class SHCloudMaskDownloader(SHDownloader):
    def __init__(self, download_folder):
        super().__init__(download_folder)
        self.resolution = 10  # mpp
        self.data_collection = DataCollection.SENTINEL2_L1C
        self.mosaicking_order = MosaickingOrder.LEAST_CC
        self.script = """
            //VERSION=3
            function setup() {
            return {
                input: ["CLM", "CLP"],
                output: { bands: 2, sampleType: "UINT8" }
            }
            }
            function evaluatePixel(sample) {
            return [ sample.CLM, sample.CLP ];
            }
            """
